@@ -29,10 +29,10 @@
 #include "gkr-keyring-item.h"
 
 #include "common/gkr-buffer.h"
-
-#include "daemon/gnome-keyring-daemon.h"
+#include "common/gkr-location.h"
 
 #include "library/gnome-keyring-memory.h"
+#include "library/gnome-keyring-private.h"
 #include "library/gnome-keyring-proto.h"
 
 #include <glib.h>
@@ -58,6 +58,12 @@ enum {
     LAST_SIGNAL
 };
 
+enum {
+    PROP_0,
+    PROP_NAME,
+    PROP_LOCATION
+};
+
 static guint signals[LAST_SIGNAL] = { 0 };
 
 G_DEFINE_TYPE (GkrKeyring, gkr_keyring, G_TYPE_OBJECT);
@@ -288,23 +294,23 @@ generate_acl_data (GkrBuffer *buffer,
 	GList *l;
 	GnomeKeyringAccessControl *ac;
 	
-	gnome_keyring_proto_add_uint32 (buffer, g_list_length (acl));
+	gkr_buffer_add_uint32 (buffer, g_list_length (acl));
 
 	for (l = acl; l != NULL; l = l->next) {
 		ac = l->data;
 		
-		gnome_keyring_proto_add_uint32 (buffer, ac->types_allowed);
-		if (!gnome_keyring_proto_add_utf8_string (buffer, ac->application->display_name)) {
+		gkr_buffer_add_uint32 (buffer, ac->types_allowed);
+		if (!gkr_proto_add_utf8_string (buffer, ac->application->display_name)) {
 			return FALSE;
 		}
-		if (!gnome_keyring_proto_add_utf8_string (buffer, ac->application->pathname)) {
+		if (!gkr_proto_add_utf8_string (buffer, ac->application->pathname)) {
 			return FALSE;
 		}
 		/* Reserved: */
-		if (!gnome_keyring_proto_add_utf8_string (buffer, NULL)) {
+		if (!gkr_proto_add_utf8_string (buffer, NULL)) {
 			return FALSE;
 		}
-		gnome_keyring_proto_add_uint32 (buffer, 0);
+		gkr_buffer_add_uint32 (buffer, 0);
 	}
 	
 	
@@ -319,28 +325,28 @@ generate_encrypted_data (GkrBuffer *buffer, GkrKeyring *keyring)
 	GkrKeyringItem *item;
 	
 	/* Make sure we're using non-pageable memory */
-	gnome_keyring_proto_go_secure (buffer);
+	gkr_proto_go_secure (buffer);
 	
 	for (l = keyring->items; l != NULL; l = l->next) {
 		item = l->data;
-		if (!gnome_keyring_proto_add_utf8_string (buffer, item->display_name)) {
+		if (!gkr_proto_add_utf8_string (buffer, item->display_name)) {
 			return FALSE;
 		}
-		if (!gnome_keyring_proto_add_utf8_secret (buffer, item->secret)) {
+		if (!gkr_proto_add_utf8_secret (buffer, item->secret)) {
 			return FALSE;
 		}
-		gnome_keyring_proto_add_time (buffer, item->ctime);
-		gnome_keyring_proto_add_time (buffer, item->mtime);
+		gkr_proto_add_time (buffer, item->ctime);
+		gkr_proto_add_time (buffer, item->mtime);
 
 		/* reserved: */
-		if (!gnome_keyring_proto_add_utf8_string (buffer, NULL)) {
+		if (!gkr_proto_add_utf8_string (buffer, NULL)) {
 			return FALSE;
 		}
 		for (i = 0; i < 4; i++) {
-			gnome_keyring_proto_add_uint32 (buffer, 0);
+			gkr_buffer_add_uint32 (buffer, 0);
 		}
 
-		if (!gnome_keyring_proto_add_attribute_list (buffer, item->attributes)) {
+		if (!gkr_proto_add_attribute_list (buffer, item->attributes)) {
 			return FALSE;
 		}
 		if (!generate_acl_data (buffer, item->acl)) {
@@ -372,38 +378,38 @@ generate_file (GkrBuffer *buffer, GkrKeyring *keyring)
 	gkr_buffer_add_byte (buffer, 0); /* crypto (0 == AEL) */
 	gkr_buffer_add_byte (buffer, 0); /* hash (0 == MD5) */
 
-	if (!gnome_keyring_proto_add_utf8_string (buffer, keyring->keyring_name)) {
+	if (!gkr_proto_add_utf8_string (buffer, keyring->keyring_name)) {
 		return FALSE;
 	}
 
-	gnome_keyring_proto_add_time (buffer, keyring->mtime);
-	gnome_keyring_proto_add_time (buffer, keyring->ctime);
+	gkr_proto_add_time (buffer, keyring->mtime);
+	gkr_proto_add_time (buffer, keyring->ctime);
 	
 	flags = 0;
 	if (keyring->lock_on_idle) {
 		flags |= 1;
 	}
-	gnome_keyring_proto_add_uint32 (buffer, flags);
-	gnome_keyring_proto_add_uint32 (buffer, keyring->lock_timeout);
-	gnome_keyring_proto_add_uint32 (buffer, keyring->hash_iterations);
+	gkr_buffer_add_uint32 (buffer, flags);
+	gkr_buffer_add_uint32 (buffer, keyring->lock_timeout);
+	gkr_buffer_add_uint32 (buffer, keyring->hash_iterations);
 	gkr_buffer_append (buffer, (guchar*)keyring->salt, 8);
 
 	/* Reserved: */
 	for (i = 0; i < 4; i++) {
-		gnome_keyring_proto_add_uint32 (buffer, 0);
+		gkr_buffer_add_uint32 (buffer, 0);
 	}
 
 	/* Hashed items: */
-	gnome_keyring_proto_add_uint32 (buffer, g_list_length (keyring->items));
+	gkr_buffer_add_uint32 (buffer, g_list_length (keyring->items));
 
 	for (l = keyring->items; l != NULL; l = l->next) {
 		item = l->data;
-		gnome_keyring_proto_add_uint32 (buffer, item->id);
-		gnome_keyring_proto_add_uint32 (buffer, item->type);
+		gkr_buffer_add_uint32 (buffer, item->id);
+		gkr_buffer_add_uint32 (buffer, item->type);
 		
-		hashed = gnome_keyring_attributes_hash (item->attributes);
+		hashed = gkr_keyring_item_attributes_hash (item->attributes);
 
-		if (!gnome_keyring_proto_add_attribute_list (buffer, hashed)) {
+		if (!gkr_proto_add_attribute_list (buffer, hashed)) {
 			gnome_keyring_attribute_list_free (hashed);
 			return FALSE;
 		}
@@ -433,7 +439,7 @@ generate_file (GkrBuffer *buffer, GkrKeyring *keyring)
 		gkr_buffer_uninit (&to_encrypt);
 		return FALSE;
 	}
-	gnome_keyring_proto_add_uint32 (buffer, to_encrypt.len);
+	gkr_buffer_add_uint32 (buffer, to_encrypt.len);
 	gkr_buffer_append (buffer, to_encrypt.buf, to_encrypt.len);
 	gkr_buffer_uninit (&to_encrypt);
 	
@@ -452,31 +458,28 @@ decode_acl (GkrBuffer *buffer, gsize offset, gsize *offset_out, GList **out)
 	
 	acl = NULL;
 
-	if (!gnome_keyring_proto_get_uint32 (buffer, offset, &offset, &num_acs)) {
+	if (!gkr_buffer_get_uint32 (buffer, offset, &offset, &num_acs)) {
 		return FALSE;
 	}
 	for (i = 0; i < num_acs; i++) {
-		if (!gnome_keyring_proto_get_uint32 (buffer, offset, &offset, &x)) {
+		if (!gkr_buffer_get_uint32 (buffer, offset, &offset, &x)) {
 			goto bail;
 		}
-		if (!gnome_keyring_proto_get_utf8_string (buffer, offset, &offset,
-							  &name)) {
+		if (!gkr_proto_get_utf8_string (buffer, offset, &offset, &name)) {
 			goto bail;
 		}
-		if (!gnome_keyring_proto_get_utf8_string (buffer, offset, &offset,
-							  &path)) {
+		if (!gkr_proto_get_utf8_string (buffer, offset, &offset, &path)) {
 			g_free (name);
 			goto bail;
 		}
-		if (!gnome_keyring_proto_get_utf8_string (buffer, offset, &offset,
-							  &reserved) ||
+		if (!gkr_proto_get_utf8_string (buffer, offset, &offset, &reserved) ||
 		    reserved != NULL) {
 			g_free (name);
 			g_free (path);
 			g_free (reserved);
 			goto bail;
 		}
-		if (!gnome_keyring_proto_get_uint32 (buffer, offset, &offset, &y)) {
+		if (!gkr_buffer_get_uint32 (buffer, offset, &offset, &y)) {
 			g_free (name);
 			g_free (path);
 			g_free (reserved);
@@ -510,7 +513,7 @@ remove_unavailable_item (gpointer key, gpointer dummy, GkrKeyring *keyring)
 	
 	g_assert (GKR_IS_KEYRING (keyring));
 	
-	item = gkr_keyring_find_item (keyring, id);
+	item = gkr_keyring_get_item (keyring, id);
 	if (item)
 		gkr_keyring_remove_item (keyring, item);
 }
@@ -537,6 +540,7 @@ update_keyring_from_data (GkrKeyring *keyring, GkrBuffer *buffer)
 	GHashTable *checks = NULL;
 	GkrKeyringItem *item;
 	char *reserved;
+	gboolean ret = FALSE;
 
 	display_name = NULL;
 	items = 0;
@@ -562,31 +566,30 @@ update_keyring_from_data (GkrKeyring *keyring, GkrBuffer *buffer)
 		return FALSE;
 	}
 
-	if (!gnome_keyring_proto_get_utf8_string (buffer, offset, &offset,
-						  &display_name)) {
+	if (!gkr_proto_get_utf8_string (buffer, offset, &offset, &display_name)) {
 		goto bail;
 	}
-	if (!gnome_keyring_proto_get_time (buffer, offset, &offset, &ctime)) {
+	if (!gkr_proto_get_time (buffer, offset, &offset, &ctime)) {
 		goto bail;
 	}
-	if (!gnome_keyring_proto_get_time (buffer, offset, &offset, &mtime)) {
+	if (!gkr_proto_get_time (buffer, offset, &offset, &mtime)) {
 		goto bail;
 	}
-	if (!gnome_keyring_proto_get_uint32 (buffer, offset, &offset, &flags)) {
+	if (!gkr_buffer_get_uint32 (buffer, offset, &offset, &flags)) {
 		goto bail;
 	}
-	if (!gnome_keyring_proto_get_uint32 (buffer, offset, &offset, &lock_timeout)) {
+	if (!gkr_buffer_get_uint32 (buffer, offset, &offset, &lock_timeout)) {
 		goto bail;
 	}
-	if (!gnome_keyring_proto_get_uint32 (buffer, offset, &offset, &hash_iterations)) {
+	if (!gkr_buffer_get_uint32 (buffer, offset, &offset, &hash_iterations)) {
 		goto bail;
 	}
-	if (!gnome_keyring_proto_get_bytes (buffer, offset, &offset, salt, 8)) {
+	if (!gkr_proto_get_bytes (buffer, offset, &offset, salt, 8)) {
 		goto bail;
 	}
 	
 	for (i = 0; i < 4; i++) {
-		if (!gnome_keyring_proto_get_uint32 (buffer, offset, &offset, &tmp)) {
+		if (!gkr_buffer_get_uint32 (buffer, offset, &offset, &tmp)) {
 			goto bail;
 		}
 		/* reserved bytes must be zero */
@@ -594,28 +597,28 @@ update_keyring_from_data (GkrKeyring *keyring, GkrBuffer *buffer)
 			goto bail;
 		}
 	}
-	if (!gnome_keyring_proto_get_uint32 (buffer, offset, &offset, &num_items)) {
+	if (!gkr_buffer_get_uint32 (buffer, offset, &offset, &num_items)) {
 		goto bail;
 	}
 
 	items = g_new0 (ItemInfo, num_items);
 
 	for (i = 0; i < num_items; i++) {
-		if (!gnome_keyring_proto_get_uint32 (buffer, offset, &offset,
+		if (!gkr_buffer_get_uint32 (buffer, offset, &offset,
 						     &items[i].id)) {
 			goto bail;
 		}
-		if (!gnome_keyring_proto_get_uint32 (buffer, offset, &offset,
+		if (!gkr_buffer_get_uint32 (buffer, offset, &offset,
 						     &items[i].type)) {
 			goto bail;
 		}
-		if (!gnome_keyring_proto_decode_attribute_list (buffer, offset, &offset,
-								&items[i].hashed_attributes)) {
+		if (!gkr_proto_decode_attribute_list (buffer, offset, &offset,
+		                                      &items[i].hashed_attributes)) {
 			goto bail;
 		}
 	}
 
-	if (!gnome_keyring_proto_get_uint32 (buffer, offset, &offset,
+	if (!gkr_buffer_get_uint32 (buffer, offset, &offset,
 					     &crypto_size)) {
 		goto bail;
 	}
@@ -635,38 +638,38 @@ update_keyring_from_data (GkrKeyring *keyring, GkrBuffer *buffer)
 			goto bail;
 		}
 		if (!verify_decrypted_buffer (&to_decrypt)) {
-			gnome_keyring_memory_free (keyring->password);
+			gnome_keyring_free_password (keyring->password);
 			keyring->password = NULL;
 		} else {
 			locked = FALSE;
 			offset += 16; /* Skip hash */
 			for (i = 0; i < num_items; i++) {
-				if (!gnome_keyring_proto_get_utf8_string (buffer, offset, &offset,
-									  &items[i].display_name)) {
+				if (!gkr_proto_get_utf8_string (buffer, offset, &offset,
+				                                &items[i].display_name)) {
 					goto bail;
 				}
-				if (!gnome_keyring_proto_get_utf8_secret (buffer, offset, &offset,
-									  &items[i].secret)) {
+				if (!gkr_proto_get_utf8_secret (buffer, offset, &offset,
+				                                &items[i].secret)) {
 					goto bail;
 				}
-				if (!gnome_keyring_proto_get_time (buffer, offset, &offset,
-								   &items[i].ctime)) {
+				if (!gkr_proto_get_time (buffer, offset, &offset,
+				                         &items[i].ctime)) {
 					goto bail;
 				}
-				if (!gnome_keyring_proto_get_time (buffer, offset, &offset,
-								   &items[i].mtime)) {
+				if (!gkr_proto_get_time (buffer, offset, &offset,
+				                         &items[i].mtime)) {
 					goto bail;
 				}
 				reserved = NULL;
-				if (!gnome_keyring_proto_get_utf8_string (buffer, offset, &offset,
-									  &reserved) ||
+				if (!gkr_proto_get_utf8_string (buffer, offset, &offset,
+				                                &reserved) ||
 				    reserved != NULL) {
 					g_free (reserved);
 					goto bail;
 				}
 				for (j = 0; j < 4; j++) {
 					guint32 tmp;
-					if (!gnome_keyring_proto_get_uint32 (buffer, offset, &offset, &tmp)) {
+					if (!gkr_buffer_get_uint32 (buffer, offset, &offset, &tmp)) {
 						goto bail;
 					}
 					/* reserved bytes must be zero */
@@ -674,8 +677,8 @@ update_keyring_from_data (GkrKeyring *keyring, GkrBuffer *buffer)
 						goto bail;
 					}
 				}
-				if (!gnome_keyring_proto_decode_attribute_list (buffer, offset, &offset,
-										&items[i].attributes)) {
+				if (!gkr_proto_decode_attribute_list (buffer, offset, &offset,
+				                                      &items[i].attributes)) {
 					goto bail;
 				}
 				
@@ -693,6 +696,7 @@ update_keyring_from_data (GkrKeyring *keyring, GkrBuffer *buffer)
 	keyring->locked = locked;
 	g_free (keyring->keyring_name);
 	keyring->keyring_name = display_name;
+	display_name = NULL;
 	keyring->mtime = mtime;
 	keyring->ctime = ctime;
 	keyring->lock_on_idle = !!(flags & LOCK_ON_IDLE_FLAG);
@@ -712,7 +716,7 @@ update_keyring_from_data (GkrKeyring *keyring, GkrBuffer *buffer)
 		/* We've seen this id */
 		g_hash_table_remove (checks, GUINT_TO_POINTER (items[i].id));
 		
-		item = gkr_keyring_find_item (keyring, items[i].id);
+		item = gkr_keyring_get_item (keyring, items[i].id);
 		if (item == NULL) {
 			item = gkr_keyring_item_new (keyring, items[i].id, items[i].type);
 			gkr_keyring_add_item (keyring, item);
@@ -723,7 +727,7 @@ update_keyring_from_data (GkrKeyring *keyring, GkrBuffer *buffer)
 
 		g_free (item->display_name);
 		item->display_name = NULL;
-		gnome_keyring_memory_free (item->secret);
+		gnome_keyring_free_password (item->secret);
 		item->secret = NULL;
 		if (item->acl) {
 			gnome_keyring_acl_free (item->acl);
@@ -748,27 +752,27 @@ update_keyring_from_data (GkrKeyring *keyring, GkrBuffer *buffer)
 	}
 	
 	g_hash_table_foreach (checks, (GHFunc)remove_unavailable_item, keyring);
-	g_hash_table_destroy (checks);
 
-	return TRUE;
+	ret = TRUE;
+
  bail:
 	gkr_buffer_uninit (&to_decrypt);
 	if (checks)
 		g_hash_table_destroy (checks);
 	g_free (display_name);
 
-	if (items != NULL) {
+	if (items != NULL && !ret) {
 		for (i = 0; i < num_items; i++) {
 			g_free (items[i].display_name);
-			gnome_keyring_memory_free (items[i].secret);
+			gnome_keyring_free_password (items[i].secret);
 			gnome_keyring_attribute_list_free (items[i].hashed_attributes);
 			gnome_keyring_attribute_list_free (items[i].attributes);
 			gnome_keyring_acl_free (items[i].acl);
 		}
-		g_free (items);
 	}
+	g_free (items);
 	
-	return FALSE;
+	return ret;
 }
 
 
@@ -793,40 +797,46 @@ write_all (int fd, const guchar *buf, size_t len)
 	return 0;
 }
 
-static char*
-get_default_keyring_file_for_name (const char *keyring_name)
+static GQuark
+get_default_location_for_name (GQuark base_loc, const char *keyring_name)
 {
-	char *base;
-	char *filename;
+	gchar *path = NULL;
+	gchar *base, *filename;
 	int version;
-	char *path;
-	char *dir;
+	GQuark loc;
+	
+	g_assert (base_loc);
+	g_assert (keyring_name && keyring_name[0]);
 
 	base = g_filename_from_utf8 (keyring_name, -1, NULL, NULL, NULL);
-	if (base == NULL) {
+	if (base == NULL)
 		base = g_strdup ("keyring");
-	}
 
-	dir = gkr_keyrings_get_dir ();
-	
 	version = 0;
 	do {
-		if (version == 0) {
-			filename = g_strdup_printf ("%s.keyring", base);
-		} else {
-			filename = g_strdup_printf ("%s%d.keyring", base, version);
-		}
+		g_free (path);
 		
-		path = g_build_filename (dir, filename, NULL);
+		if (version == 0) 
+			filename = g_strdup_printf ("%s/keyrings/%s.keyring", 
+			                            g_quark_to_string (base_loc), base);
+		else
+			filename = g_strdup_printf ("%s/keyrings/%s%d.keyring", 
+			                            g_quark_to_string (base_loc), base, version);
+
+		loc = gkr_location_from_string (filename);
 		g_free (filename);
+		
+		path = gkr_location_to_path (loc);
+		g_return_val_if_fail (path, 0);
 
 		version++;
 	} while (g_file_test (path, G_FILE_TEST_EXISTS));
 
 	g_free (base);
-	g_free (dir);
 	
-	return path;
+	loc = gkr_location_from_path (path);
+	g_free (path);
+	return loc;
 }
 
 /* -----------------------------------------------------------------------------
@@ -846,6 +856,22 @@ gkr_keyring_init (GkrKeyring *keyring)
 	init_salt (keyring->salt);
 }
 
+static void
+gkr_keyring_get_property (GObject *obj, guint prop_id, GValue *value, 
+                          GParamSpec *pspec)
+{
+	GkrKeyring *keyring = GKR_KEYRING (obj);
+
+	switch (prop_id) {
+	case PROP_NAME:
+		g_value_set_string (value, keyring->keyring_name);
+		break;
+	case PROP_LOCATION:
+		g_value_set_uint (value, keyring->location);
+		break;
+	}
+}
+
 static void 
 gkr_keyring_dispose (GObject *obj)
 {
@@ -862,7 +888,7 @@ gkr_keyring_dispose (GObject *obj)
 	g_list_free (keyring->items);
 	keyring->items = NULL;
 	
-	gnome_keyring_memory_free (keyring->password);
+	gnome_keyring_free_password (keyring->password);
 	keyring->password = NULL;
 
 	G_OBJECT_CLASS (gkr_keyring_parent_class)->dispose (obj);
@@ -874,7 +900,6 @@ gkr_keyring_finalize (GObject *obj)
 	GkrKeyring *keyring = GKR_KEYRING (obj);
 
 	g_free (keyring->keyring_name);
-	g_free (keyring->file);
 	g_assert (keyring->password == NULL);
 	
 	G_OBJECT_CLASS (gkr_keyring_parent_class)->finalize (obj);
@@ -887,9 +912,18 @@ gkr_keyring_class_init (GkrKeyringClass *klass)
 
 	gkr_keyring_parent_class  = g_type_class_peek_parent (klass);
 	
+	gobject_class->get_property = gkr_keyring_get_property;
 	gobject_class->dispose = gkr_keyring_dispose;
 	gobject_class->finalize = gkr_keyring_finalize;
 	
+	g_object_class_install_property (gobject_class, PROP_NAME,
+		g_param_spec_string ("name", "Name", "Keyring Name",
+		                     NULL, G_PARAM_READABLE));
+		                     
+	g_object_class_install_property (gobject_class, PROP_LOCATION,
+		g_param_spec_uint ("location", "Location", "File Location",
+		                   0, G_MAXUINT, 0, G_PARAM_READABLE));
+	
 	signals[ITEM_ADDED] = g_signal_new ("item-added", GKR_TYPE_KEYRING, 
 			G_SIGNAL_RUN_FIRST, G_STRUCT_OFFSET (GkrKeyringClass, item_added),
 			NULL, NULL, g_cclosure_marshal_VOID__OBJECT, 
@@ -906,7 +940,7 @@ gkr_keyring_class_init (GkrKeyringClass *klass)
  */
 
 GkrKeyring*
-gkr_keyring_new (const char *name, const char *path)
+gkr_keyring_new (const char *name, GQuark location)
 {
 	GkrKeyring *keyring;
 	
@@ -915,19 +949,25 @@ gkr_keyring_new (const char *name, const char *path)
 	keyring = g_object_new (GKR_TYPE_KEYRING, NULL);
 	
 	keyring->keyring_name = g_strdup (name);
-	keyring->file = g_strdup (path);
+	keyring->location = location;
 
 	return keyring;
 }
 
 GkrKeyring*
-gkr_keyring_create (const gchar *keyring_name, const gchar *password)
+gkr_keyring_create (GQuark base_loc, const gchar *keyring_name, const gchar *password)
 {
 	GkrKeyring *keyring;
+
+	g_return_val_if_fail (keyring_name && keyring_name[0], NULL);
+	g_return_val_if_fail (password, NULL);
+		
+	if (!base_loc)
+		base_loc = GKR_LOCATION_BASE_LOCAL;
 	
-	keyring = gkr_keyring_new (keyring_name, NULL);
+	keyring = gkr_keyring_new (keyring_name, 0);
 	if (keyring != NULL) {
-		keyring->file = get_default_keyring_file_for_name (keyring_name);
+		keyring->location = get_default_location_for_name (base_loc, keyring_name);
 		keyring->locked = FALSE;
 		keyring->password = gnome_keyring_memory_strdup (password);
 		gkr_keyring_save_to_disk (keyring);
@@ -959,7 +999,7 @@ gkr_keyring_get_new_id (GkrKeyring *keyring)
 }
 
 GkrKeyringItem*
-gkr_keyring_find_item (GkrKeyring *keyring, guint id)
+gkr_keyring_get_item (GkrKeyring *keyring, guint id)
 {
 	GkrKeyringItem *item;
 	GList *l;
@@ -973,6 +1013,22 @@ gkr_keyring_find_item (GkrKeyring *keyring, guint id)
 	return NULL;
 }
 
+GkrKeyringItem*  
+gkr_keyring_find_item (GkrKeyring *keyring, GnomeKeyringItemType type, 
+                       GnomeKeyringAttributeList *attrs)
+{    
+	GkrKeyringItem *item;
+	GList *l;
+	
+	for (l = keyring->items; l; l = g_list_next (l)) {
+		item = GKR_KEYRING_ITEM (l->data);
+		if (gkr_keyring_item_match (item, type, attrs, TRUE))
+			return item;
+	}
+	
+	return NULL;
+}
+
 void
 gkr_keyring_add_item (GkrKeyring* keyring, GkrKeyringItem* item)
 {
@@ -1016,39 +1072,58 @@ gkr_keyring_update_from_disk (GkrKeyring *keyring, gboolean force_reload)
 	char *contents = NULL;
 	gsize len;
 	gboolean success = FALSE;
+	gchar *file = NULL;
 
-	if (keyring->file == NULL)
-		return TRUE;
+	if (!keyring->location) {
+		success = TRUE;
+		goto done;
+	}
 
-	if (stat (keyring->file, &statbuf) < 0)
+	file = gkr_location_to_path (keyring->location);
+	if (!file)
+		goto done;
+		
+	if (stat (file, &statbuf) < 0) {
+		g_free (file);
 		return FALSE;
+	}
 
-	if (!force_reload && statbuf.st_mtime == keyring->file_mtime)
-		return TRUE;
+	if (!force_reload && statbuf.st_mtime == keyring->file_mtime) {
+		success = TRUE;
+		goto done;
+	}
 	keyring->file_mtime = statbuf.st_mtime;
 
-	if (!g_file_get_contents (keyring->file, &contents, &len, NULL))
-		return FALSE;
-
-	gkr_buffer_init_static (&buffer, (guchar*)contents, len);
+	if (!g_file_get_contents (file, &contents, &len, NULL))
+		goto done;
 	
+	gkr_buffer_init_static (&buffer, (guchar*)contents, len);
 	success = update_keyring_from_data (keyring, &buffer);
 	gkr_buffer_uninit (&buffer);
-	g_free (contents);
 
+done:
+	g_free (contents);
+	g_free (file);
 	return success;
 }
 
 gboolean 
 gkr_keyring_remove_from_disk (GkrKeyring *keyring)
 {
+	gchar *file;
 	int res;
 
 	/* Cannot remove session or memory based keyring */
-	if (keyring->file == NULL)
+	if (!keyring->location)
 		return FALSE;
-
-	res = unlink (keyring->file);
+		
+	file = gkr_location_to_path (keyring->location);
+	if (!file)
+		return FALSE;
+		
+	res = unlink (file);
+	g_free (file);
+	
 	return (res == 0);
 }
 
@@ -1061,21 +1136,26 @@ gkr_keyring_save_to_disk (GkrKeyring *keyring)
 	char *dirname;
 	char *template;
 	gboolean ret = TRUE;
+	gchar *file = NULL;
 	
-	if (keyring->locked) {
-		/* Can't save locked keyrings */
+	/* Can't save locked keyrings */
+	if (keyring->locked)
 		return FALSE;
-	}
 
-	if (keyring->file == NULL) {
-		/* Not file backed */
+	/* Not file backed */
+	if (!keyring->location)
 		return TRUE;
-	}
+		
+	file = gkr_location_to_path (keyring->location);
+	if (!file)
+		return FALSE;
 	
 	gkr_buffer_init_full (&out, 4096, g_realloc);
 
 	if (generate_file (&out, keyring)) {
-		dirname = g_path_get_dirname (keyring->file);
+		dirname = g_path_get_dirname (file);
+		if (g_mkdir_with_parents (dirname, S_IRWXU) < 0)
+			g_warning ("unable to create keyring dir");
 		template = g_build_filename (dirname, ".keyringXXXXXX", NULL);
 		
 		fd = g_mkstemp (template);
@@ -1086,10 +1166,9 @@ gkr_keyring_save_to_disk (GkrKeyring *keyring)
 			fsync (fd);
 #endif
 				close (fd);
-				if (rename (template, keyring->file) == 0) {
-					if (stat (keyring->file, &statbuf) == 0) {
+				if (rename (template, file) == 0) {
+					if (stat (file, &statbuf) == 0)
 						keyring->file_mtime = statbuf.st_mtime;
-					}
 				} else {
 					unlink (template);
 				}
@@ -1109,6 +1188,51 @@ gkr_keyring_save_to_disk (GkrKeyring *keyring)
 	}
 	
 	gkr_buffer_uninit (&out);
+	g_free (file);
 	return ret;
 }
 
+gboolean
+gkr_keyring_lock (GkrKeyring *keyring)
+{
+	if (keyring->locked)
+		return TRUE;
+
+	/* Never lock the session keyring */
+	if (!keyring->location)
+		return TRUE;
+
+	g_assert (keyring->password != NULL);
+	
+	gnome_keyring_free_password (keyring->password);
+	keyring->password = NULL;
+	if (!gkr_keyring_update_from_disk (keyring, TRUE)) {
+		/* Failed to re-read, remove the keyring */
+		g_warning ("Couldn't re-read keyring %s\n", keyring->keyring_name);
+		gkr_keyrings_remove (keyring);
+	}
+	
+	return TRUE;
+}
+
+gboolean
+gkr_keyring_unlock (GkrKeyring *keyring, const gchar *password)
+{
+	if (!keyring->locked)
+		return TRUE;
+		
+	g_assert (keyring->password == NULL);
+		
+	keyring->password = gnome_keyring_memory_strdup (password);
+	if (!gkr_keyring_update_from_disk (keyring, TRUE)) {
+		gnome_keyring_free_password (keyring->password);
+		keyring->password = NULL;
+	}
+	if (keyring->locked) {
+		g_assert (keyring->password == NULL);
+		return FALSE;
+	} else {
+		g_assert (keyring->password != NULL);
+		return TRUE;
+	}
+}
